/* VERSION 3: VIDEO WITH TEXT CHAT, DISPLAY WHAT IS SENT, INSTEAD OF WHAT IS CAPTURED.  
 * =====================================================
 * Last updated: June 29, 2007.
 * 
 * Features:
 * - Simple two-way video conferencing 
 * - Text Chat
 * - Mirror image displays what is being captured, actual image displays what is being sent (what the other side will see).
 * - videoDisplay shows what is actually being sent; size of videoDisplay matters. (if bigger, there is more of a delay time)
 * - (interface changes to show mirror display and actual display)
 */


using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Text;
using System.Windows.Forms;

namespace VideoChat
{
    public partial class Form1 : Form
    {
        // variable declarations       
        bool present, recording;
        int myLocation;
        static int LOCATION_SCHOOL = 1;
        static int LOCATION_HOME = 2;
        EasyImages.CameraClient wc;
        GroupLab.GTNetworking.SharedDictionary SD;        
        GroupLab.GTNetworking.Subscription imageSubscription;
        GroupLab.GTNetworking.Subscription pingSubscription;
        GroupLab.GTNetworking.Subscription textSubscription; 
        GroupLab.GTNetworking.Subscription isHereSubscription;
        GroupLab.GTNetworking.Subscription waitingSubscription;
        Timer awarenessTimer;
                       
        /* Thread handling: Need this block of code to make sure code runs from the main windows form, and not any other thread */
        public delegate void SetControlImage(PictureBox icontrol, Bitmap image); // delegate: pointer to a function
        public void RecievedImageToControl(PictureBox icontrol, Image image)
        {
            /* Control.InvokeRequired -- gets a value indicating whether the caller must call an invoke method when making method 
             * calls to the control because the caller is on a different thread than the one the control was created on. 
             * Control.BeginInvoke() -- executes a delegate asynchronously on the thread that the control's underlying handle was created on.
             */             
            if (icontrol.InvokeRequired) 
            {
                SetControlImage theDelegate = new SetControlImage(RecievedImageToControl);
                this.BeginInvoke(theDelegate, new object[] { icontrol, image });                
            }
            else
                icontrol.Image = image;
        }//RecievedImageToControl

        public Form1()
        {
            Control.CheckForIllegalCrossThreadCalls = false;
            InitializeComponent();

            // camera declarations
            wc = new EasyImages.CameraClient("DefaultCamera");            
            wc.ReceivedFrame += new EasyImages.CameraClient.ReceivedFrameEventHandler(wc_ReceivedFrame);
            wc.Start();
            present = false;
            recording = false;
            // shared dictionary and subscription declarations
            SD = new GroupLab.GTNetworking.SharedDictionary();
            pingSubscription = new GroupLab.GTNetworking.Subscription();
            imageSubscription = new GroupLab.GTNetworking.Subscription();
            textSubscription = new GroupLab.GTNetworking.Subscription();
            isHereSubscription = new GroupLab.GTNetworking.Subscription();
            waitingSubscription = new GroupLab.GTNetworking.Subscription();

            // Timer declaration
            awarenessTimer = new Timer();
            awarenessTimer.Interval = 1000;
            awarenessTimer.Tick += new EventHandler(awarenessTimer_Tick);
            
        }

        private void awarenessTimer_Tick(object sender, EventArgs e)
        {
            takeSnap();
        }//Form1

        private void wc_ReceivedFrame(object sender, EasyImages.CameraEventArgs e)
        {
            RecievedImageToControl(this.mirrorDisplay, e.Bitmap); //this.MirrorDisplay.Image = e.Bitmap;
            
        }//client_ReceivedFrame

        private void ConnectionButton_Click(object sender, EventArgs e)
        {
            if (this.connectionButton.Text == "Connect!")
            {
                string url = "tcp://localhost:hello";

                if (DialogResult.OK == GroupLab.GTNetworking.UrlPrompt.Show(ref url))
                {
                    this.SD.Url = url;                    
                    this.SD.Opened += new EventHandler(SD_Opened);
                    this.SD.Closed +=  new EventHandler(SD_Closed);
                    this.SD.Open(); // start server or client
                    this.connectionButton.Text = "Disconnect!";                  
                }
            }
            else
            {
                this.connectionButton.Text = "Connect!";
                this.SD.Close(); // close the connection with the shared dictionary
            }
        }//ConnectionButton_Click

        private void SD_Opened(object sender, EventArgs e)
        {
            // imageSubscription
            this.imageSubscription.BeginInit();
            this.imageSubscription.Dictionary = this.SD;
            if (this.SD.ConnectionType == GroupLab.GTNetworking.ConnectionType.Client)
                this.imageSubscription.Pattern = "/ServerImage";
            else
                this.imageSubscription.Pattern = "/ClientImage";
            this.imageSubscription.Notified += new GroupLab.GTNetworking.SubscriptionEventHandler(imageSubscription_Notified);
            this.imageSubscription.EndInit();

            // pingSubscription
            this.pingSubscription.BeginInit();
            this.pingSubscription.Dictionary = this.SD;
            if (this.SD.ConnectionType == GroupLab.GTNetworking.ConnectionType.Client)
                this.pingSubscription.Pattern = "/ServerPing";
            else
                this.pingSubscription.Pattern = "/ClientPing";            
            this.pingSubscription.Notified += new GroupLab.GTNetworking.SubscriptionEventHandler(pingSubscription_Notified);
            this.pingSubscription.EndInit();

            // send the first frame
            //takeSnap();

            // textSubscription
            this.textSubscription.BeginInit();
            this.textSubscription.Dictionary = this.SD;
            if (this.SD.ConnectionType == GroupLab.GTNetworking.ConnectionType.Client)
                this.textSubscription.Pattern = "/ServerTextMessage";
            else
                this.textSubscription.Pattern = "/ClientTextMessage";            
            this.textSubscription.Notified += new GroupLab.GTNetworking.SubscriptionEventHandler(textSubscription_Notified);
            this.textSubscription.EndInit();


            // isHereSubscription
            this.isHereSubscription.BeginInit();
            this.isHereSubscription.Dictionary = this.SD;
            if (this.SD.ConnectionType == GroupLab.GTNetworking.ConnectionType.Client)
                this.isHereSubscription.Pattern = "/ServerIsHere";
            else
            {
                this.isHereSubscription.Pattern = "/ClientIsHere";
                this.SD["/ClientIsHere"] = false;
                this.SD["/ServerIsHere"] = false;
            }
            this.isHereSubscription.Notified +=new GroupLab.GTNetworking.SubscriptionEventHandler(isHereSubscription_Notified);
            this.isHereSubscription.EndInit();
            
            if (this.SD.ConnectionType == GroupLab.GTNetworking.ConnectionType.Client)
            {
                myLocation = LOCATION_HOME;
                this.waitingSubscription.BeginInit();
                this.waitingSubscription.Dictionary = this.SD;
                this.waitingSubscription.Pattern = "/WaitingOnHome";
                this.waitingSubscription.Notified += new GroupLab.GTNetworking.SubscriptionEventHandler(waitingSubscription_Notified);
                this.waitingSubscription.EndInit();
            }
            else
            {
                myLocation = LOCATION_SCHOOL;
            }

            this.interfaceKit1.SensorChange += new GroupLab.SharedPhidgets.IndexSensorChangeEventHandler(this.interfaceKit1_SensorChange);
            awarenessTimer.Start();
        }




        private void SD_Closed(object sender, System.EventArgs args)
        {
        }//SD_Closed

        private void takeSnap()
        {
            if (this.SD.Status == GroupLab.GTNetworking.ConnectionStatus.Opened)
            {
                if (this.SD.ConnectionType == GroupLab.GTNetworking.ConnectionType.Client)
                {
                    this.locationLabel.ForeColor = Color.Navy;
                    this.locationLabel.Text = "Location: HOME";
                    this.SD["/ClientImage"] = new EasyImages.JPEG(wc.GetFrame(), 39);
                }
                else if (this.SD.ConnectionType == GroupLab.GTNetworking.ConnectionType.Server)
                {
                    this.locationLabel.ForeColor = Color.DarkGreen;                    
                    this.locationLabel.Text = "Location: SCHOOL";
                    this.SD["/ServerImage"] = new EasyImages.JPEG(wc.GetFrame(), 39);
                }

            }//if

        }//takeSnap

        void imageSubscription_Notified(object sender, GroupLab.GTNetworking.SubscriptionEventArgs e)//imageSubscription_Notified
        {
            EasyImages.JPEG j = e.Value as EasyImages.JPEG;
            if (j != null)
            {               
                RecievedImageToControl(this.videoDisplay, j.GetImage());  //this.VideoDisplay.Image = j.GetImage();                    
            }
            if (this.SD.ConnectionType == GroupLab.GTNetworking.ConnectionType.Client)
            {
                // Display what is actually sent over network
                if (this.SD.Contains("/ClientImage"))
                    RecievedImageToControl(this.actualDisplay, (this.SD["/ClientImage"] as EasyImages.JPEG).GetImage()); 
            
                // indicate that image has been recieved and can now get the next one (stop and wait)
                this.SD["/ClientPing"] = "ping!"; 
            }
            else if (this.SD.ConnectionType == GroupLab.GTNetworking.ConnectionType.Server)
            {
                // Display what is actually sent over network
                 if (this.SD.Contains("/ServerImage"))                    
                    RecievedImageToControl(this.actualDisplay, (this.SD["/ServerImage"] as EasyImages.JPEG).GetImage());

                // indicate that image has been recieved and can now get the next one (stop and wait)
                this.SD["/ServerPing"] = "ping!";
            }
        }//imageSubscription_Notified

        private void pingSubscription_Notified(object sender, GroupLab.GTNetworking.SubscriptionEventArgs e)
        {
            if(recording)takeSnap(); //if got, then send the next frame
        }//pingSubscription_Notified

        void waitingSubscription_Notified(object sender, GroupLab.GTNetworking.SubscriptionEventArgs e)
        {
            this.interfaceKit1.Outputs[0].State = (bool) e.Value;
            // ALERT SAUL
        }

        private void textSubscription_Notified(object sender, GroupLab.GTNetworking.SubscriptionEventArgs e)
        {
            this.textLabel.Text = e.Value.ToString();
        }//textSubscription_Notified


        private void CheckPresence()
        {
            bool otherPresent = SD.ConnectionType.Equals(GroupLab.GTNetworking.ConnectionType.Client) ? (bool)this.SD["/ServerIsHere"] : (bool)this.SD["/ClientIsHere"];
            // case 1: This is here, other isn't
            if (present && !otherPresent)
            {
                if (myLocation == LOCATION_HOME)
                {
                    // home here, school not here -- send 1 frame per second, basic awareness on both sides

                }
                else if (myLocation == LOCATION_SCHOOL)
                {
                    // school here, home not here
                    this.SD["/WaitingOnHome"] = true;
                }
            }
            // case 2: This isn't here, other is  
            else if (!present && otherPresent)
            {
                if (myLocation == LOCATION_HOME)
                {
                    // home not here, school here
                    // this.SD["/WaitingOnHome"] = true;
                }
                else
                {
                    // school not here, home here -- send 1 frame per second, basic awareness on both sides
                }
            }
            // case 3: both there
            else if (present && otherPresent)
            {
                recording = true;
                awarenessTimer.Stop();
                takeSnap();
            }
            // case 4: both gone
            else
            {
                // send 1 frame per second, basic awareness on both sides
                recording = false;
                awarenessTimer.Start();
            }
        }

        private void isHereSubscription_Notified(object sender, GroupLab.GTNetworking.SubscriptionEventArgs e)
        {

            CheckPresence();
        }//isHereSubscription_Notified


        private void Form1_FormClosing(object sender, FormClosingEventArgs e)
        {
            // if user chooses to close the form manually, do so in an elegant fashion (stop camera and close dictionary)
            wc.Stop();
            if (SD.Status == GroupLab.GTNetworking.ConnectionStatus.Opened)
            {
                SD.Close();
            }
        }//Form1_FormClosing

        private void chatBox_KeyPress(object sender, KeyPressEventArgs e)
        {
            if (this.SD.ConnectionType == GroupLab.GTNetworking.ConnectionType.Client)
            {
                this.SD["/ClientTextMessage"] = chatTextBox.Text + e.KeyChar.ToString();
            }
            else if (this.SD.ConnectionType == GroupLab.GTNetworking.ConnectionType.Server)
            {
                this.SD["/ServerTextMessage"] = chatTextBox.Text + e.KeyChar.ToString();
            }
            if (e.KeyChar == '\r')
            {
                chatTextBox.Text = "";
            }
        }//chatBox_KeyPress

        private void button1_Click(object sender, EventArgs e)
        {
            if (this.button1.Text.Equals("+"))
            {
                this.videoDisplay.Size = new System.Drawing.Size(320, 240);
                this.button1.Text = "-";

            }
            else if (this.button1.Text.Equals("-"))
            {
                this.videoDisplay.Size = new System.Drawing.Size(160, 120);
                this.button1.Text = "+";
            }//else
        }

        private void interfaceKit1_SensorChange(object sender, GroupLab.SharedPhidgets.IndexSensorChangeEventArgs e)
        {
            if (SD.Status != GroupLab.GTNetworking.ConnectionStatus.Opened) return;

            if (e.Index == 0)   // Light sensor on chair to detect presence
            {                
                if (e.Value < 100)
                {
                    present = true;
                    if (this.SD.ConnectionType == GroupLab.GTNetworking.ConnectionType.Client)
                    {
                        this.SD["/ClientIsHere"] = true;
                    }
                    else if (this.SD.ConnectionType == GroupLab.GTNetworking.ConnectionType.Server)
                    {
                        this.SD["/ServerIsHere"] = true;
                    }
                }
                else if (e.Value >= 100)
                {
                    present = false;
                    recording = false;
                    awarenessTimer.Start();
                    if (this.SD.ConnectionType == GroupLab.GTNetworking.ConnectionType.Client)
                    {
                        this.SD["/ClientIsHere"] = false;
                        
                    }
                    else if (this.SD.ConnectionType == GroupLab.GTNetworking.ConnectionType.Server)
                    {
                        this.SD["/ServerIsHere"] = false;
                        this.SD["/WaitingOnHome"] = false; 
                    }                 
                }

            }//if light sensor
            CheckPresence();
        }//button1_Click

     
    }// partial class Form1
}// namespace VideoChat